#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include "webminifier.h"


char *load_file (FILE *input);
void write_file (char *content, FILE *output);


int main (int argc, char *argv[]) {
	FILE *input;
	FILE *output;
	char *ext;
	char *content;
	char *minified;
	
	if (argc != 3) {
		printf ("usage: %s /path/to/source /path/to/minified\n", argv[0]);
		exit (1);
	}
	ext = strrchr (argv[1], '.');
	if (ext == NULL || (strcasecmp (".html", ext) != 0 && strcasecmp (".css", ext) != 0 && strcasecmp (".js", ext) != 0)) {
		printf ("error: unsupported file type\n");
		exit (2);
	}
	
	input = fopen (argv[1], "r");
	if (input == NULL) {
		printf ("error: can't read %s\n", argv[1]);
		exit (3);
	}
	output = fopen (argv[2], "w");
	if (input == NULL) {
		printf ("error: can't write %s\n", argv[2]);
		fclose (input);
		exit (4);
	}
		
	content = load_file (input);
	minified = (char *) malloc (strlen (content) * sizeof (char));
	if (content == NULL || minified == NULL) {
		printf ("error: could not allocate memory\n");
		if (content != NULL) free (content);
		fclose (input);
		fclose (output);
		remove (argv[2]);
		exit(5);
	}

	if (strcasecmp (".html", ext) == 0) {
		minify_html (content, minified);
	} else if (strcasecmp (".css", ext) == 0) {
		minify_css (content, minified, strlen (content));
	} else if (strcasecmp (".js", ext) == 0) {
		minify_js (content, minified, strlen (content));
	}
	printf ("%lu - %lu => %lu bytes (%2.0f%%)\n", strlen (content), 
												  strlen (content) - strlen (minified), 
												  strlen (minified), 
												  (float) 100.0 * strlen (minified) / strlen (content));
	write_file (minified, output);
	free (content);
	free (minified);
	fclose (input);
	fclose (output);
}


char *load_file (FILE *input) {
	fseek (input, 0, SEEK_END);
	long size = ftell (input);
	fseek (input, 0, SEEK_SET);
	char *content = (char *) malloc ((size + 1) * sizeof (char));
	if (content != NULL) {
		long offset = 0;
		while (offset != size) {
			offset += fread (content + offset, sizeof (char), size - offset, input);
		}
	}
	content[size] = '\0';
	return content;
}


void write_file (char *content, FILE *output) {
	long size = strlen(content);
	long offset = 0;
	while (offset != size) {
		offset += fwrite (content + offset, sizeof(char), size - offset, output);
	}
}
