#include <string.h>
#include <stdbool.h>
#include <stdint.h>
#include <regex.h>
#include "webminifier.h"


long accept (char *input, char *start, char *end);
	/**
	 * searches start and end in input string
	 * example: long size = accept(input, "<!--", "-->");
	 * @param input string to search
	 * @param start string
	 * @param end string
	 * @return lenght of the string or -1 if not found
	 */

long copy_tag (char *input, char *output, long size);
	/**
	 * copy HTML tag
	 * @param input HTML string
	 * precondition: input[0] == '<'
	 * @param output (i/o) minified string
	 * @param size "<tag ...>" input string lenght
	 * @return lenght of the minified output string
	 */
	 
bool is_spc (char c);
	/** tells if c is a space (' '), tab ('\t') or newline ('\n'or '\r')
	 * @param c character
	 * @return true or false
	 */
	 
bool replace_str (char *input, char *output, long *input_idx, long *output_idx, const char *in_str, const char *out_str);
	/** replace in_str in input string (from *input_idx) by out_str in output 
	 * (from *output_idx) if (input + *input_idx) starts with in_str <=> if 
	 * strcmp(in_str, input + *input_idx) == 0
	 * @param input string
	 * @param output (i/o) string
	 * @param input_idx (i/o) the start of input string - if replacement 
	 * is made, input_idx is incremented by strlen (in_str).
	 * @param outut_idx (i/o) the start of output string - if replacement
	 * is made, output_idx is incremented by strlen (out_str).
	 * @return true if replacement is made
	 */


void minify_html (char *input, char *output) {
	long input_idx = 0;
	long output_idx = 0;
	long size = strlen (input);
	long part_size;
	int c, p;
	
	while (input_idx != size) {
		part_size = accept (input + input_idx, "<!--", "-->"); //skip comments
		if (part_size != -1) {
			input_idx += part_size;
			continue;
		}
		
		part_size = accept (input + input_idx, "<pre>", "</pre>"); //keep format
		if (part_size != -1) {
			strncpy (output + output_idx, input + input_idx, part_size);
			input_idx += part_size;
			output_idx += part_size;
			continue;
		}
		
		part_size = accept (input + input_idx, "<style", "</style>");
		if (part_size != -1) {
			while (input[input_idx] != '>') { input_idx++; part_size--; }
			strcpy (output + output_idx, "<style>");
			output_idx += 7; //strlen("<style>")
			output_idx += minify_css(input + input_idx + 1L, output + output_idx, part_size - (1 + 8) /*strlen("</style>")*/);
			strcpy (output + output_idx, "</style>");
			output_idx += 8;
			input_idx += part_size;
		}
		//TODO: <script>
		
		part_size = accept (input + input_idx, "<", ">"); //Element
		if (part_size != -1) {
			output_idx += copy_tag (input + input_idx, output + output_idx, part_size);
			input_idx += part_size;
			continue;
		}
		
		//TextNode
		if (is_spc (input[input_idx])) {
			while (is_spc (input[input_idx + 1])) input_idx++; //skip multiple spaces
		}
		p = c; //remember last character
		c = input[input_idx];
		if (is_spc (c)) c = ' ';
		if (c != ' ' || (p != ' ' && !(output[output_idx - 1] == '>' && input[input_idx + 1] == '<'))) { //only keep text and spaces between words
			output[output_idx] = c;
			output_idx++;
		}
		input_idx++;
	}
	
	output[output_idx] = '\0';
}


long minify_css (char *input, char *output, long size) {
	const char *separators = ",:{};>~^[]|\"'/!"; // '/' but not '*' (selector), '-' and '+' (calc)
	long input_idx = 0;
	long output_idx = 0;
	char c, p, n;
	long part_size;
	regex_t zero_regex; regcomp(&zero_regex, "^[: ] *0 *[ep][xtm]", 0); //(em|pt|px)
	regmatch_t zero_match[1];
	
	while (input_idx != size) {
		part_size = accept (input + input_idx, "/*", "*/"); //skip comments
		if (part_size != -1) {
			input_idx += part_size;
			continue;
		}
		
		if (replace_str (input, output, &input_idx, &output_idx, "black", "#000")) continue;
		if (replace_str (input, output, &input_idx, &output_idx, "white-space", "white-space")) continue;
		if (replace_str (input, output, &input_idx, &output_idx, "white", "#fff")) continue;
		if (replace_str (input, output, &input_idx, &output_idx, "bolder", "bolder")) continue;
		if (replace_str (input, output, &input_idx, &output_idx, "bold", "700")) continue;
		if (replace_str (input, output, &input_idx, &output_idx, "normal", "400")) continue;
		if (replace_str (input, output, &input_idx, &output_idx, ":nth-child(1)", ":first-child")) continue;
		
		if (regexec(&zero_regex, input + input_idx, 1, zero_match, 0) == 0) {
			output[output_idx] = input[input_idx];
			output[output_idx + 1] = '0';
			output_idx += 2;
			input_idx += (intmax_t) zero_match[0].rm_eo;
			continue;
		}
		
		if (is_spc (input[input_idx])) {
			while (is_spc (input[input_idx + 1])) input_idx++; //skip multiple spaces
		}
		
		p = output[output_idx - 1];
		c = input[input_idx];
		n = input[input_idx + 1];
		if (is_spc (c)) c = ' ';
		if (c != ' ' || (strchr (separators, p) == NULL && strchr (separators, n) == NULL)) { //if not a space or not near a separator
			if (c == '}' && p == ';') {
				output[output_idx - 1] = '}'; //remove ';' before '}'
			} else {
				output[output_idx] = c;
				output_idx++;
			}
		}
		
		if (!(p >= 0x30 && p <= 0x39) && c == '0' && n == '.') output_idx--; // 0.123 -> .123
		
		input_idx++;
	}
	output[output_idx] = '\0';
	return output_idx;
}


long minify_js (char *input, char *output, long size) {
	strcpy (output, input);
	return strlen (output);
}


long accept(char *input, char *start, char *end) {
	long idx = 0;
	uint8_t startlen = strlen(start);
	uint8_t endidx = 0;
	uint8_t endlen = strlen(end);
	bool valid = true;
	bool accepted = false;
	
	while (valid && !accepted && input[idx] != '\0') {
		if (idx < startlen) {
			if (input[idx] != start[idx]) valid = false;
		} else {
			//TODO ' / "
			if (input[idx] != end[endidx]) {
				endidx = 0;
			} else {
				endidx++;
				if (endidx == endlen) {
					accepted = true;
				}
			}
		}
		idx++;
	}
	if (!accepted) idx = -1;
	return idx;
}


long copy_tag (char *input, char *output, long size) {
	long input_idx;
	long output_idx = 0;
	for (input_idx = 0; input_idx < size; input_idx++) {
		if (!is_spc (input[input_idx]) || (input[input_idx + 1] != '/' && input[input_idx + 1] != '>' && !is_spc (output[output_idx -1]))) { //don't copy spaces before '/' or '>'
			if (is_spc (input[input_idx])) output[output_idx] = ' ';
			else output[output_idx] = input[input_idx];
			output_idx++;
		}
	}
	return output_idx;
}


bool is_spc (char c) {
	return c == ' ' || c == '\n' || c == '\t' || c == '\r';
}


bool replace_str (char *input, char *output, long *input_idx, long *output_idx, const char *in_str, const char *out_str) {
	if (strncmp (input + *input_idx, in_str, strlen (in_str)) == 0) {
		strcpy (output + *output_idx, out_str);
		*input_idx += strlen (in_str);
		*output_idx += strlen (out_str);
		return true;
	}
	return false;
}
